!******************************************************************************
!
! File : jdqz_interior.f90
!   by : Alan Tackett
!   on : 04/17/00
!  for : 
!
!
!
!******************************************************************************



!******************************************************************************
! JDQZ_Interior - Jacobi-Davidson Band solver
!  Uses B orthogonal vectors
!******************************************************************************

Subroutine JDQZ_interior(Target_shift, eigen, eivec, &
     ClusterInfo, EigIn, n, shift, eps, &
     kmax, jmax, jmin, Qmax, method, m, L, maxnmv, maxstep, &
     lock, order, work, lwork,out_unit,UseRandom)
   Use jd_misc
   Real,    Intent(IN)    :: Target_Shift
   Real,    Intent(INOUT) :: Eigen(:)
   Complex, Intent(INOUT) :: eivec(:,:)
   Integer, Intent(IN)    :: ClusterInfo(:,:)
   Integer, Intent(IN)    :: EigIn
   Integer, Intent(IN)    :: N
   Complex, Intent(IN)    :: Shift
   Real,    Intent(IN)    :: eps
   Integer, Intent(INOUT) :: kmax
   Integer, Intent(IN)    :: jmax
   Integer, Intent(IN)    :: jmin
   Integer, Intent(IN)    :: Qmax
   Integer, Intent(IN)    :: method
   Integer, Intent(IN)    :: m
   Integer, Intent(IN)    :: L
   Integer, Intent(INOUT) :: maxnmv
   Integer, Intent(INOUT) :: maxstep
   Real,    Intent(IN)    :: lock
   Integer, Intent(IN)    :: order
   Complex, Target, Intent(OUT)   :: Work(:,:)
   Integer, Intent(INOUT) :: Lwork
   Integer, Intent(IN)    :: out_unit
   Logical, Intent(IN)    :: UseRandom

   Integer :: testspace, totalmv
   Complex :: alpha(jmax), beta(jmax)
   logical :: loop, try, found, ok
   integer :: ldvs, ldzwork, iseed(4),s,t, mode
   Complex :: ma(jmax,jmax), mb(jmax,jmax), zma(jmax,jmax), zmb(jmax,jmax)
   Complex :: vsl(jmax,jmax), vsr(jmax,jmax), ra(jmax,jmax), rb(jmax,jmax)
   Complex :: Gvsl(jmax,jmax), Gvsr(jmax,jmax)
!   Complex :: zwork(8*jmax), aconv(jmax), bconv(jmax)
   Complex :: zwork(4*jmax*jmax+7*jmax+1)
   integer :: ldqkz, copied, ncluster, clusterCopied, ClusterFound
   integer :: ipivqkz(jmax), NumToDrop, NumToKeep
   Complex :: mqkz(jmax,jmax), invqkz(jmax,jmax), f(jmax), dot

   integer :: i, j, k, info, mxmv, step, Qpos, Ktotal, bw, aq, zindex
   integer :: d, u, v, w, aux, av, bv, q, z, kz, itmp, tp, bq
   integer :: solvestep, LUT(jmax)
   Real    :: rnrm, rwork(8*Jmax), deps, dtmp, reig(jmax), Emin, tol
   Real    :: NewTarget, currerr, reig2(jmax)
   Complex :: zalpha, zbeta, targeta, targetb, evcond, shifta, shiftb
   Complex :: cdot, Wdot(Jmax)
   Complex :: dznrm2      !** External function 

   Integer :: CopyPos, NumFound, totalprec, totalgmresmv

   common /debug2/ idbg(20), outunit
   integer idbg, outunit
   integer :: kstr(4,4*jmax+24)
   Integer :: Index(8), struc(100), pstruc(4)
   Real    :: epsu, gap, adelta, bdelta
   Logical :: drop_zero, forceStore, FirstBand

   Complex, PARAMETER :: zero = (0,0)
   Complex, PARAMETER :: one = (1,0)
   data iseed /3,3,1966,29/            

   mode = 1
Write(*,*) 'JDQZ_WBS: Mode=',mode

   ldvs = jmax;  ldzwork = 8*jmax
   ldqkz = jmax
   ldzwork = 4*ldvs*ldvs + 7*ldvs + 1   + 4*ldvs

   !** Are there errors in the parameters?
!*   if ( kmax .gt. jmax ) call error('jdqz: kmax greater than jmax')
   if ( method .ne. 1 .and. method .ne. 2 ) &
        call error('jdqz: illegal choice for solution method')
   if ( order .lt. -2 .or. order .gt. 2 ) &
        call error('illegal value for order, must be between -2 and 2')

   d   = 1  !** d= rhs, these ptr's refer to the columns of the workspace
   tp  = d+1 !** Workspace for jdqzmv
   u   = tp+1 !** u   = pointer to Krylov space GMRES(m) or Bi-CSTAB(l)
   !**  v   = pointer to search space JDQZ with max dimension jmax
   if ( method .eq. JD_gmres ) then
      v   = u+m+1
   else if ( method .eq. JD_cgstab ) then
      v   = u+2*l+6
   end if
   w   = v+jmax !** pointer to test subspace JDQZ with max dimension jmax
   bw   = w+jmax !** pointer to test subspace JDQZ with max dimension jmax
   av  = bw+jmax !** pointer to subspace AV with max dimension jmax
   bv  = av+jmax  !** pointer to subspace BV with max dimension jmax
   aux = bv+jmax
   q   = aux+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
   aq  = q+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
   bq  = aq+jmax !** ptr to search Schur basis in JDQZ with max dim kmax
!   wq  = bq+jmax
   z   = bq+kmax  !** pointer to test Schur basis in JDQZ with max dim kmax
   kz  = z+kmax  !** kz  = pointer to matrix K^{-1}Z_k

   if (lwork <= 0) then   !** Return best value for Lwork
      lwork = kz+Kmax - 1
      RETURN
   End If

   if (kz+kmax-1.gt.lwork) then
      Write(out_unit,*) 'LWORK=',lwork, ' * should be ',kz+kmax-1
      call error ('qz: memory fault')
   End If

   !*** initialize loop **

   testspace = 3
   ok = .true.

   NewTarget = shift
   evcond = dcmplx(sqrt(abs(shift)**2+abs(one)**2))
   shifta = shift/evcond
   shiftb = one/evcond   
!   shifta = shift
!   shiftb = 1.0

   targeta = shifta
   targetb = shiftb
   Emin = -1E10

   zalpha = shifta
   zbeta = shiftb

   step = 0
   deps = dble(one)
   mxmv = 0

   solvestep = 0
   copied = 0
   NumFound = 0
   Qpos = 0
   CopyPos = 0
   Ktotal = 0

   j = 0
   k = 0
   totalmv = 0
   totalprec = 0
   totalgmresmv = 0
   ncluster = 1
   NumFound = Min(Jmin-1, ClusterInfo(1,ncluster))
   ClusterFound = 0
   ClusterCopied = 0

Write(*,*) 'Jmin=',Jmin, ' * Jmax=',Jmax, ' * Kmax=',Kmax, ' * Qmax=',Qmax, ' * MaxNMV=',MaxNMV
   !** Main Loop
   Loop = .TRUE.
   Do While (Loop)
      step = step+1
      solvestep = solvestep+1
      write(*,*) 'Outer step=',step, ' * SolveStep=',SolveStep, ' * j=',j, ' * copied=',copied, ' * EigIn=',EigIn, ' * K=',k, ' * Ktotal=',Ktotal, ' * NF=',NumFound
      Write(*,*) '     * a,b=',zalpha,zbeta
      Write(*,*) '     * eig=',(zalpha/zbeta)
      Write(*,*) '     sa,sb=',Shifta, shiftb, ' * sa/sb=',ShiftA/ShiftB
      Write(*,*) '     ta,tb=',Targeta, Targetb, ' * ta/tb=',Targeta/Targetb

      mxmv = 0
      if (j==0) then
         call zlarnv(2, iseed, n, work(1,v+j))
         call zlarnv(2, iseed, n, work(1,w+j))
         do i=1,n
            dtmp = dble(work(i,v+j))
            work(i,v+j) = dcmplx(dtmp,0d0)
            dtmp = dble(work(i,w+j))
            work(i,w+j) = dcmplx(dtmp,0d0)
         enddo
      else 
         deps = 4d0**(-solvestep)
!deps = eps
         if ((NumFound > 0) .AND. (copied<EigIn)) then
            copied = copied + 1
            Work(:,v+j) = EiVec(:,copied)           
            zalpha = Eigen(copied); zbeta = 1; 
            NumFound = NumFound - 1
            ClusterCopied = ClusterCopied + 1
Write(*,*) 'Adding CC=',ClusterCopied

            SolveStep = 0

Write(*,*) 'JDQZ: AFTER FOUND j=',j, ' * copied=',copied, ' * SolveStep=',solvestep, ' * NumFound=',numfound, ' * Eigen=',REAL(Zalpha)
         else if (j.lt.jmin) then
Write(*,*) 'JDQZ:  ZGMRES INIT DDOT=',DOT_PRODUCT(Work(:,d), Work(:,d))
               mxmv = 2
!               SolveStep = 0
!               mxmv = maxnmv
Write(*,*) 'mxmv=',mxmv, ' * j,k=',j,k, ' * eps=',eps, ' * m=',m, ' za.zb=',zalpha,zbeta
               call zgmres(n, work(1,v+j), work(1,d), m, deps, &
                    mxmv, zalpha, zbeta, k+1, &
                    work(1,kz), work(1,q), invqkz, ldqkz, &
                    ipivqkz, f, work(1,u), work(1,tp) )

Write(*,*) 'VDOT=',DOT_PRODUCT(Work(:,v+j),Work(:,v+j))              
!Write(*,*) 'JDQZ: AFTER FOUND j=',j, ' * copied=',copied, ' * SolveStep=',solvestep, ' * NumFound=',numfound           
         else if (method.eq.JD_gmres) then
Write(*,*) 'JDQZ:  ZGMRES DDOT=',DOT_PRODUCT(Work(:,d), Work(:,d))
Write(*,*) 'JDQZ:  ZGMRES ZDOT=',DOT_PRODUCT(Work(:,z+k), Work(:,z+k))
Write(*,*) 'JDQZ:  ZGMRES KZDOT=',DOT_PRODUCT(Work(:,kz+k), Work(:,kz+k))
            mxmv = maxnmv
            call zgmres(n, work(1,v+j), work(1,d), m, deps, &
                 mxmv, zalpha, zbeta, k+1, &
                 work(1,kz), work(1,q), invqkz, ldqkz, &
                 ipivqkz, f, work(1,u), work(1,tp) )
         else if (method.eq.JD_cgstab) then
Write(*,*) 'JDQZ:  ZCGSTABL'
            mxmv = maxnmv
            call zcgstabl(n, work(1,v+j), work(1,d), l, &
                 deps, mxmv, zalpha, &
                 zbeta, k+1, work(1,kz), work(1,q), invqkz, &
                 ldqkz, ipivqkz, f, work(1,u), 2*l+6)
         end if      
      end if
      
      TotalMV = TotalMV + mxmv
      TotalGMRESmv = TotalGMRESmv + mxmv
Write(*,*) 'step=',step, ' * TotalGMRESMV=',TotalGMRESMV

Write(*,*) '######## VDOT=',Dot_PRODUCT(Work(:,v+j),Work(:,v+j))
!Write(*,*) '######## DDOT=',Dot_PRODUCT(Work(:,d),Work(:,d))

      j = j+1

      call zmgs (n, j-1, work(1,v), work(1,v+j-1), 1)
      call zmgs (n, k, work(1,q), work(1,v+j-1), 1)

Write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! j,k,copied=',j,k,copied

      Call ABmul(n, work(1,v+j-1), work(1,av+j-1), work(1,bv+j-1))
      totalmv = totalmv + 1

      !*** Form new W and Orthonormalize ***
      Work(:,w+j-1) = shiftb*Work(:,av+j-1) - shifta*Work(:,bv+j-1)
      call zmgs (n, j-1, work(1,w), work(1,w+j-1), 1)
      call zmgs (n, k, work(1,z), work(1,w+j-1), 1)      

      !** FIX ME!!!
!**      Call ABmul(n, work(1,v+j-1), work(1,av+j-1), work(1,bv+j-1))

      call makemm (n, j, work(1,w), work(1,av), ma, zma, ldvs)
      call makemm (n, j, work(1,w), work(1,bv), mb, zmb, ldvs)


      try = .true.
      if ((j<Jmin) .AND. .NOT. UseRandom) try = .FALSE.
      If ((.NOT. Try) .AND. (copied >= EigIn)) Try = .TRUE.

!Write(*,*) 'Ma----------------------------------------------'
!Do i=1, j
!   Write(*,*) 'JDQZ_WBS: i=',i, ' * ',Ma(i,1:j)
!End do
!Write(*,*) 'Mb----------------------------------------------'
!Do i=1, j
!   Write(*,*) 'JDQZ_WBS: i=',i, ' * ',Mb(i,1:j)
!End do

try=.true.   !** FIXME **
      If (try) then
!         call zgegs ('v', 'v', j, zma, ldvs, zmb, ldvs, &
!             alpha, beta, vsl, ldvs, vsr, ldvs, zwork, &
!             ldzwork, rwork, info)

         epsu = 1E-10
         gap = 1000
         drop_zero = .FALSE.
         t = j
         Call GUPTRI(zMa(1,1), zMb(1,1), ldvs, t, t, epsu, gap, drop_zero, &
              VSL(1,1), ldvs, VSR(1,1), ldvs, &
              adelta, bdelta, index(1), index(2), index(3), index(4), &
              index(5), index(6), index(7), index(8), &
              pstruc(1), struc(1), zwork(1), kstr(1,1), info)
         
         i = index(8)-Index(2)
         If (i /= t) then
            Write(*,*) 'JDQZ_WBS: GUPTRI INCE-RTCE=',i, ' * Should be',t
            Write(*,*) 'JDQZ_WBS: Index=',Index
            STOP
         End If
         If (Info /= 0) then
            Write(*,*) 'JD_WindowBandSolver: GUPTRI =',info
            STOP
         End If

         Do i=1, j
            alpha(i) = zMa(i,i)
            beta(i) = zMb(i,i)
            reig(i) = alpha(i)/beta(i)
         End do
Write(*,*) 'Raw eig=', reig(1:j), ' * TARGET=',REAL(targeta/targetb)
      End If


      Do While (try)
         !*** Sort the Petrov pairs ***
         Call QZSort(targeta, targetb, j, zma, zmb, vsl, vsr, ldvs, &
                 alpha, beta, order)

         Do i=1, j
            reig(i) = alpha(i)/beta(i)
         End do
Write(*,*) 'After QZSORT!!!!! eig=', reig(1:j), ' * TARGET=',REAL(targeta/targetb)

         zalpha = zma(1,1);  zbeta = zmb(1,1);         
         evcond = dcmplx(sqrt(abs(zalpha)**2+abs(zbeta)**2))
         
         !*** compute new q ***
         call zgemv ('n', n, j, one, work(1,v), n, vsr(1,1), &
              1, zero, work(1,q+k), 1)
         call zmgs (n, k, work(1,q), work(1,q+k), 1)

write(*,*) 'Q k+1=',k+1, ' DOT=',DOT_PRODUCT(Work(:,q+k),Work(:,q+k))
         !*** Calc new Z ***
         call zgemv ('n', n, j, one, work(1,w), n, vsl(1,1), &
              1, zero, work(1,z+k), 1)
         call zmgs (n, k, work(1,z), work(1,z+k), 1 )
         
         
         !*** Make new qkz ***
         call zcopy (n, work(1,z+k), 1, work(1,kz+k), 1)
         call precon (n, work(1,kz+k))
         totalprec = totalprec + 1
         call mkqkz (n, k+1, work(1,q), work(1,kz), mqkz, invqkz, &
              ldqkz, ipivqkz)

         !*** compute new (right) residual= beta Aq - alpha Bq and
         !*** orthogonalize this vector on Z.
         Call ABmul(n, work(1,q+k), work(1,aq+k), work(1,bq+k))
         Work(:,d) = zbeta*Work(:,aq+k) - zalpha*Work(:,bq+k)
         call zmgs (n, k, work(1,z), work(1,d), 0 )

         rnrm = dznrm2 (n, work(1,d), 1)/dble(evcond)

currerr = rnrm
Write(*,*) 'CURRERR=',rnrm
Write(*,*) 'DDOT=',Dot_PRODUCT(Work(:,d),Work(:,d))


         if (rnrm.lt.lock.and.ok) then
            targeta = zalpha
            targetb = zbeta
            ok = .false.
         end if
         
         found = (rnrm.lt.eps.and.(j.gt.1.or.k.eq.kmax-1))
         try =  found
        
         Write(Out_Unit,*) 'JDQZI:', step, totalmv, currerr, k, REAL(zalpha/zbeta), found
         if (found) then              
            FirstBand = .FALSE.
            SolveStep = 0

            !*** increase the number of found evs by 1 ***
            k = k+1
            Ktotal = Ktotal + 1
            Qpos = Qpos + 1
            ClusterFound = ClusterFound + 1
Write(*,*) 'FOUND START CF=',ClusterFound, ' * CC=',ClusterCopied
Write(*,*) '   NC=',Ncluster, ' * CI=',ClusterInfo(:,ncluster), ' * NF=',NumFound
            If (.NOT. USeRandom) then
               If (Clusterfound == ClusterInfo(1,ncluster)) then
                  ncluster = ncluster + 1
                  NumFound = Min(Jmin-1, ClusterInfo(1,ncluster))
                  NumToDrop = ClusterInfo(2,ncluster)
                  NumToKeep = k - NumToDrop
                  ClusterFound = 0
                  ClusterCopied = 0
               else
                  NumToDrop = 0
                  NumToKeep = k
                  if (ClusterCopied < ClusterInfo(1,ncluster)) then
                     NumFound = NumFound + 1
                  End if
               End if
            else
               NumtoDrop = 0; NumtoKeep = 0;
            End If
Write(*,*) 'FOUND END CF=',ClusterFound, ' * CC=',ClusterCopied
Write(*,*) '   NC=',Ncluster, ' * CI=',ClusterInfo(:,ncluster)
Write(*,*) '   NTD=',NumtoDrop, ' * NTK=',NumToKeep, ' * NF=',NumFound

write(*,*) 'FOUND step=',step, ' * k=',k, ' * eig=',REAL(zalpha/zbeta), ' * CalcEig=',REAL(cdot), ' * ERR=',currerr, ' * CF=',ClusterFound, ' *nc=',ncluster, ' * CI(:,nc)=',Clusterinfo(:,ncluster), ' * Ktotal=',Ktotal, ' * Kmax=',Kmax
            
            !*** store the eigenvalue
            Eigen(ktotal) = zalpha/zbeta

            solvestep = 0

            If (ktotal < Kmax) then
               call zgemm ('n', 'n', n, j-1, j, one, work(1,v), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = v
               v = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,av), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = av
               av = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,bv), n, &
                    vsr(1,2), ldvs, zero, work(1,aux), n)
               itmp = bv
               bv = aux
               aux = itmp
               call zgemm ('n', 'n', n, j-1, j, one, work(1,w), n, &
                    vsl(1,2), ldvs, zero, work(1,aux), n)
               itmp = w
               w = aux
               aux = itmp

               j = j-1
               call zlacpy ('a', j, j, zma(2,2), ldvs, ma, ldvs)
               call zlacpy ('a', j, j, ma, ldvs, zma, ldvs)
               call zlacpy ('a', j, j, zmb(2,2), ldvs, mb, ldvs)
               call zlacpy ('a', j, j, mb, ldvs, zmb, ldvs)
               call zlaset ('a', j, j, zero, one, vsr, ldvs)
               call zlaset ('a', j, j, zero, one, vsl, ldvs)
               targeta = shifta
               targetb = shiftb
               ok = .true.
               mxmv = 0
               deps = dble(one)
            else
               try = .FALSE.
            end IF
         else if (j >= jmax) then
Write(*,*) 'SHRINKING!!!!!!!!!!!!!!!!!!!!!!!!!'
            call zgemm ('n', 'n', n, jmin, j, one, work(1,v), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = v
            v = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,av), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = av
            av = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,bv), n, &
                 vsr, ldvs, zero, work(1,aux), n)
            itmp = bv
            bv = aux
            aux = itmp
            call zgemm ('n', 'n', n, jmin, j, one, work(1,w), n, &
                 vsl, ldvs, zero, work(1,aux), n)
            itmp = w
            w = aux
            aux = itmp
            
            j = jmin
            call zlacpy ('a', j, j, zma, ldvs, ma, ldvs)
            call zlacpy ('a', j, j, zmb, ldvs, mb, ldvs)
            call zlaset ('a', j, j, zero, one, vsr, ldvs)
            call zlaset ('a', j, j, zero, one, vsl, ldvs)
         End If
      End Do   !** TRY **
      
      loop = (ktotal.lt.kmax.and.step.lt.maxstep)
   End Do  !** End Big Loop ***
   

   !** Did enough eigenpairs converge?
   kmax = ktotal


   !** Form the Schur matrices 
   call zlaset ('l', k, k, zero, zero, ra, ldvs)    
   call zlaset ('l', k, k, zero, zero, rb, ldvs)    
   do i = 1, k
      call ABmul( n, work(1,q+i-1), work(1,tp), work(1,d) )
      totalmv = totalmv + 1
      call zgemv( 'c', n, i, one, work(1,z), n, work(1,tp), 1,  &
           zero, ra(1,i), 1 )
      call zgemv( 'c', n, i, one, work(1,z), n, work(1,d), 1,  &
           zero, rb(1,i), 1 )
   end do

   !*** Now calculate the eigenvectors and eigenvalues
   call zgegv('N','V',k,ra, ldvs, rb, ldvs, alpha, beta,vsl, ldvs, &
                   vsr,ldvs, zwork,ldzwork,rwork,info)

   !*** Finally store the eigenvectors and eigenvalues
   call zgemm('n', 'n', n, k, k, one, work(1,q), n, &
        vsr, ldvs, zero, eivec, n)

   Eigen(1:Ktotal) = alpha(1:ktotal)/beta(1:ktotal)
   maxnmv = totalmv
   Write(*,*) 'JDQZ_BS: FOUND TotalMV = ',TotalMV, ' * TotalPREC = ', TotalPrec
   Write(*,*) 'JDQZ_BS: FOUND TotalGMRESMV = ',TotalGMRESMV
   Write(*,*) 'JDQZ_BS: FOUND Eigen=',Eigen(1:kTotal)
 
   Return
End Subroutine
